package util;

import org.blinkenlights.jid3.ID3Exception;
import org.blinkenlights.jid3.MediaFile;
import org.blinkenlights.jid3.MP3File;
import org.blinkenlights.jid3.v2.ID3V2_3_0Tag;
import org.blinkenlights.jid3.v2.APICID3V2Frame;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.*;
import java.util.*;

import enums.FileType;
import properties.DefaultProperties;

public class Mp3Processor {

    private static String ERR_TMP_NOT_FOUND = "unable to find file make sure you have Pandora running in a FIRFOX browser";

    private static Log LOG = LogFactory.getLog(Mp3Processor.class);

    public static File findCurrentMP3() throws IOException {
        LOG.info("locating mp3");
        File currentTmpMp3 = findCurrentTmpMp3();
        LOG.debug("checking file integrity for file: " + currentTmpMp3.getName());
        int loopCount = 0;
        while (true) {
            try {
                LOG.debug("sleeping for a second");
                Thread.sleep(1000);

            } catch (InterruptedException e) {
                //ignore
            }
            File currentMp3Refetched = findCurrentTmpMp3();
            if (currentMp3Refetched.getName().equals(currentTmpMp3.getName())) {
                LOG.debug("file integrity check complete");
                break;
            }
            if (loopCount > 40) {
                throw new RuntimeException("File integrity could not be confirmed after 40 loops - process aborted");
            }
            loopCount++;
        }

        return copyFile(currentTmpMp3, new File(currentTmpMp3.getParent(), "pandorasJarTmpFile.mp3"));
    }

    private static File findCurrentTmpMp3() {
        int retryCount = DefaultProperties.getTmpMp3RetryCount();
        File[] mp3Files = new File[0];
        for (int x = 0; x < retryCount; x++) {
            try {
                mp3Files = getJoinedPandoraMp3FilesInDecendingOrder();
                break;
            } catch (Exception e) {
                if (x + 1 == retryCount) {
                    throw new RuntimeException(e.getMessage());
                }
                LOG.info("Mp3 file was not located retrying after 1 second. loop count: " + (x +1));
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e1) {
                    //ignore
                }
            }
        }
        return mp3Files[1];
    }


    private static File[] getJoinedPandoraMp3FilesInDecendingOrder() throws IOException {
        List<File> mp3s = new ArrayList<File>();
        File tempDir = findTempDir();
        joinMp3s(mp3s, tempDir);
        File[] joinedMp3s = mp3s.toArray(new File[mp3s.size()]);
        sortFilesByLastModified(joinedMp3s);
        if (joinedMp3s.length < 2) {
            throw new RuntimeException(ERR_TMP_NOT_FOUND);
        }
        return joinedMp3s;
    }

    private static File findTempDir() {
        String tmpDir = System.getenv("TEMP");
        if (tmpDir == null || tmpDir.length() == 0) {
            tmpDir = System.getenv("TMP");
        }
        if (tmpDir == null) {
            throw new RuntimeException(ERR_TMP_NOT_FOUND);
        }
        return new File(tmpDir);
    }

    private static boolean isFileRecent(File file) {
        Date now = new Date();
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(new Date(file.lastModified()));
        calendar.add(Calendar.HOUR, 3);
        return now.before(calendar.getTime());
    }

    public static void joinMp3s(List<File> mp3s, File lookinDir) throws IOException {
        LOG.debug("\n***processing directory: " + lookinDir.getName() + "***\n");
        File[] files = lookinDir.listFiles();
        sortFilesByLastModified(files);

        //uncomment to print files in each working directory
        //printFiles(files);

        for (File currentFile : files) {
            if (!currentFile.isDirectory()) {
                if (isMp3(currentFile)) {
                    LOG.debug("mp3 found adding currentFile: " + currentFile.getName());
                    mp3s.add(currentFile);
                } else {
                    if (isFileRecent(currentFile)) {
                        LOG.debug("recent non mp3 file found: " + currentFile.getName() + " - will continue to next file");
                        continue;
                    }
                    LOG.debug("no more mp3 found in dir");
                    return;
                }
            } else {
                joinMp3s(mp3s, currentFile);
            }
        }
        LOG.debug("no more mp3 found in dir");
    }

    private static void printFiles(File[] files) {
        LOG.debug("working with files");
        for (File file : files) {
            LOG.debug(file.getName());
        }
    }

    private static boolean isMp3(File file) {
        FileInputStream is = null;
        try {
            is = new FileInputStream(file);
            byte[] buffer = new byte[32 * 1024];
            while (is.read(buffer) != -1) {
                String fragment = new String(buffer);
                if (fragment.indexOf("LAME") != -1) {
                    return true;
                }
            }
            LOG.debug("file " + file.getName() + " is not an mp3");
            return false;
        }
        catch (IOException e) {
            LOG.debug("unable to parse file: " + file + " marking file as non-mp3");
            return false;
        }
        finally {
            if (is != null) {
                try {
                    is.close();
                }
                catch (IOException e) {
                    LOG.info("unable to clode input stream: " + e.getMessage());
                }
            }
        }
    }


    public static void sortFilesByLastModified(File[] files) {
        Arrays.sort(files, new Comparator() {
            public int compare(Object o1, Object o2) {
                if (((File) o1).lastModified() > ((File) o2).lastModified()) {
                    return -1;
                } else if (((File) o1).lastModified() < ((File) o2).lastModified()) {
                    return +1;
                } else {
                    return 0;
                }
            }
        });
    }

    public static File copyFile(File src, File dst) throws IOException {
        LOG.info("copying file[" + src.getAbsolutePath() + "] to temp[" + dst.getAbsolutePath() + "]");
        InputStream in = new FileInputStream(src);
        OutputStream out = new FileOutputStream(dst);
        byte[] buf = new byte[1024];
        int len;
        while ((len = in.read(buf)) > 0) {
            out.write(buf, 0, len);
        }
        in.close();
        out.close();
        return dst;
    }

    public static void addID3Tags(File currentMP3, File coverArt, String comment, SongInfo songInfo) throws ID3Exception, IOException {
        LOG.info("adding ID3 tags: " + songInfo.toString());
        MediaFile oMediaFile = new MP3File(currentMP3);
        ID3V2_3_0Tag oID3V2_3_0Tag = new ID3V2_3_0Tag();

        if (coverArt != null) {
            LOG.info("setting ID3 cover art");
            byte[] buffer = new byte[32 * 1024];
            BufferedInputStream bufferedInputStream = new BufferedInputStream(new FileInputStream(coverArt));
            ByteArrayOutputStream imageBinary = new ByteArrayOutputStream();
            while (bufferedInputStream.read(buffer) != -1) {
                imageBinary.write(buffer);
            }
            APICID3V2Frame apicid3V2Frame = new APICID3V2Frame("image/jpg", APICID3V2Frame.PictureType.Artist, songInfo.getAlbum() != null ? songInfo.getAlbum() : "unknown", imageBinary.toByteArray());
            oID3V2_3_0Tag.addAPICFrame(apicid3V2Frame);
        }
        if (songInfo.getAlbum() != null) {
            oID3V2_3_0Tag.setAlbum(songInfo.getAlbum());
        }
        if (songInfo.getArtist() != null) {
            oID3V2_3_0Tag.setArtist(songInfo.getArtist());
        }
        if (comment != null) {
            if (songInfo.getStationName() != null) {
                oID3V2_3_0Tag.setComment(songInfo.getStationName());
            } else {
                oID3V2_3_0Tag.setComment(comment);
            }
        }
        if (songInfo.getTrackNumber() != null) {
            Integer i = new Integer(songInfo.getTrackNumber());
            oID3V2_3_0Tag.setTrackNumber(i);
        }
        if (songInfo.getGenre() != null) {
            oID3V2_3_0Tag.setGenre(songInfo.getGenre());
        }
        if (songInfo.getTitle() != null) {
            oID3V2_3_0Tag.setTitle(songInfo.getTitle());
        }
        if (songInfo.getYear() != null) {
            try {
                Integer y = new Integer(songInfo.getYear());
                oID3V2_3_0Tag.setYear(y);
            } catch (NumberFormatException e) {
                //ignore
            }

        }

        oMediaFile.setID3Tag(oID3V2_3_0Tag);
        oMediaFile.sync();
    }

    public static void saveMP3(SongInfo songInfo, File fileToSave, FileType fileType, DataOutputStream output, boolean addToItunes) throws IOException {
        File file = new File(Util.createSongHierarchy(songInfo), SafeString.getSafeFileName(songInfo.getArtist().trim() + "-" + songInfo.getTitle().trim()) + "." + fileType.toString().toLowerCase());
        Mp3Processor.copyFile(fileToSave, file);
        String mp3Path = file.getAbsolutePath();
        if (addToItunes) {
            ITunesManager.addTrackToItunes(mp3Path, songInfo.getStationName());
        }
        LOG.info("\nripped file to: " + mp3Path + "\nyea baby!!!\n");
        if (fileType == FileType.MP3) {
            output.writeBytes("<span style=\"font: 12px arial\">MP3 tagged and saved:</span><br/><span style=\"font: 12px arial\">" + mp3Path + "</span>");
        }
    }
}
